#!/usr/bin/env python3

import os

import requests

from .error import ParameterError, ResourceNotFound


class CloudSigmaClient(object):
    def __init__(self, region=None, username=None, password=None):

        region = region or os.getenv("CLOUDSIGMA_REGION")
        username = username or os.getenv("CLOUDSIGMA_USERNAME")
        password = password or os.getenv("CLOUDSIGMA_PASSWORD")

        # use hack to pass credentials directly to pycloudsigma module
        # See github issue:
        #   https://github.com/cloudsigma/pycloudsigma/issues/7
        # Initialization workaround adapted from:
        #   http://blog.sourcepole.ch/2016/11/22/manually-initializing-the-pycloudsigma-module/
        # now import (and initialize) the cloudsigma module

        # allow pycloudsigma to read a null config file
        os.putenv("CLOUDSIGMA_CONFIG", "/dev/null")
        import cloudsigma

        # monkeypatch the config values into the module config
        cloudsigma.conf.config.__setitem__(
            "api_endpoint", f"https://{region}.cloudsigma.com/api/2.0/"
        )
        cloudsigma.conf.config.__setitem__(
            "ws_endpoint", f"wss://direct.{region}.cloudsigma.com/websocket"
        )
        cloudsigma.conf.config.__setitem__("username", username)
        cloudsigma.conf.config.__setitem__("password", password)

        # save config values for local image upload
        self.username = username
        self.password = password
        self.upload_endpoint = (
            f"https://direct.{region}.cloudsigma.com/api/2.0/drives/upload/"
        )

        self.config = cloudsigma.conf.config
        self.server = cloudsigma.resource.Server()
        self.drive = cloudsigma.resource.Drive()
        self.vlan = cloudsigma.resource.VLAN()
        self.ip = cloudsigma.resource.IP()
        self.subscription = cloudsigma.resource.Subscriptions()
        self.capabilities = cloudsigma.resource.Capabilites()

        self.list_format = None

    def _get_name(self, uuid, _type):
        if _type == "server":
            server = self.find_server(uuid)
            name = server.get("name")
        elif _type == "drive":
            drive = self.find_drive(uuid)
            name = drive.get("name")
        elif _type == "vlan":
            vlan = self.find_vlan(uuid)
            name = vlan["meta"].get("name")
        elif _type == "ip":
            ip = self.find_ip(uuid)
            name = ip["meta"].get("name")
        elif _type == "subscription":
            name = None
        else:
            raise ParameterError("unknown resource type {_type}")
        return name or f"<unnamed_{_type}>"

    def _format_resource(self, resource, item, list_format):
        if resource == self.server:
            label = "server"
            count = item["smp"]
            speed = int(item["cpu"]) / count
            nics = []
            for nic in item["nics"]:
                mac = nic["mac"]
                if nic["vlan"]:
                    config = "vlan"
                else:
                    config = nic["ip_v4_conf"]["conf"]
                if config == "static":
                    if nic["runtime"]:
                        ip = nic["runtime"]["ip_v4"]["uuid"]
                    else:
                        ip = nic["ip_v4_conf"]["ip"]["uuid"]
                elif config == "dhcp":
                    if nic["runtime"]:
                        ip = nic["runtime"]["ip_v4"]["uuid"]
                    else:
                        ip = "<assigned-on-boot>"
                elif config in ["manual", "vlan"]:
                    ip = "<os-configured>"
                else:
                    raise ParameterError(f"unknown nic conf value {config}")
                nics.append({mac: dict(config=config, ip=ip)})
            if item["cpus_instead_of_cores"]:
                smp_type = "cpu"
            else:
                smp_type = "core"
            data = [
                dict(name=self._get_name(item["uuid"], label)),
                dict(status=item["status"]),
                dict(cpu=f"{count}"),
                dict(clock=f"{speed/1000}Ghz"),
                dict(smp=smp_type),
                dict(memory=self.format_memory_value(item["mem"])),
                dict(
                    drives=[
                        self._get_name(drive["drive"]["uuid"], "drive")
                        for drive in item["drives"]
                    ]
                ),
                dict(nics=nics),
            ]
        elif resource == self.drive:
            label = "drive"
            mounted = item["mounted_on"]
            if len(mounted):
                mounted_on = [
                    self._get_name(server["uuid"], "server") for server in mounted
                ]
            else:
                mounted_on = []
            data = [
                dict(name=self._get_name(item["uuid"], label)),
                dict(size=self.format_memory_value(item["size"])),
                dict(media=item["media"]),
                dict(storage_type=item["storage_type"]),
                dict(mounted=mounted_on),
            ]
        elif resource == self.vlan:
            label = "vlan"
            data = [
                dict(name=self._get_name(item["uuid"], label)),
                dict(description=item["meta"].get("description")),
            ]
        elif resource == self.ip:
            label = "ip"
            data = [
                dict(name=self._get_name(item["uuid"], label)),
                dict(
                    server=[
                        self._get_name(item["server"]["uuid"], "server")
                        if item["server"]
                        else "unassigned"
                    ]
                ),
                dict(description=item["meta"].get("description")),
            ]
        elif resource == self.subscription:
            label = "subscription"
            data = [dict(name=self._get_name(item["uuid"], label)), dict(detail="")]
        else:
            raise ParameterError(f"Unknown resource: {resource}")

        if list_format == "text":
            dlines = [""]
            for dd in data:
                assert (len(dd)) == 1
                for k, v in dd.items():
                    if isinstance(v, list):
                        if k == "nics":
                            dlines.append(f"{k}=[")
                            for vv in v:
                                dlines.append(f"  {vv}")
                            dlines.append("]")
                        elif k == "drives":
                            dlines.append(f"{k}={v}")
                    else:
                        if len(dlines[-1]):
                            dlines[-1] += "  "
                        dlines[-1] += f"{k}={v}"
            longest = max([len(dd) for dd in dlines])
            dlines = [d + (" " * (longest - len(d))) for d in dlines][:longest]

            data = dlines

        return {item["uuid"]: data}

    def _list_resources(self, resource, list_format):
        if resource not in (self.subscription, self.capabilities) and list_format:
            resources = resource.list_detail()
        else:
            resources = resource.list()

        if list_format == "uuid":
            resources = [{i["uuid"]: None} for i in resources]
        elif list_format in ["brief", "text"]:
            resources = [
                self._format_resource(resource, item, list_format) for item in resources
            ]
        elif list_format in ["detail", None]:
            # no post-processing
            pass
        else:
            raise ParameterError(f"unknown list_format {self.list_format}")

        return resources

    def list_all(self, list_format):
        resources = {}
        all_resources = dict(
            servers=self.server, drives=self.drive, vlans=self.vlan, ips=self.ip
        )
        for label, resource in all_resources.items():
            resources[label] = self._list_resources(resource, list_format)
        return resources

    def list_servers(self, list_format):
        return dict(servers=self._list_resources(self.server, list_format))

    def list_drives(self, list_format):
        return dict(drives=self._list_resources(self.drive, list_format))

    def list_vlans(self, list_format):
        return dict(vlans=self._list_resources(self.vlan, list_format))

    def list_ips(self, list_format):
        return dict(ips=self._list_resources(self.ip, list_format))

    def list_subscriptions(self, list_format):
        if list_format not in ["uuid", "detail"]:
            list_format = "detail"
        return dict(subscriptions=self._list_resources(self.subscription, list_format))

    def list_capabilities(self, list_format):
        return dict(capabilities=[self._list_resources(self.capabilities, "detail")])

    def _find_resource(self, resource_lister, _type, name):
        for resource in list(resource_lister("detail").values())[0]:
            if name in [resource.get("name"), resource.get("uuid")]:
                return resource
        raise ResourceNotFound(f"unknown {_type} {name}")

    def find_server(self, name):
        return self._find_resource(self.list_servers, "server", name)

    def find_drive(self, name):
        return self._find_resource(self.list_drives, "drive", name)

    def find_vlan(self, name):
        return self._find_resource(self.list_vlans, "vlan", name)

    def find_ip(self, name=None):
        return self._find_resource(self.list_ips, "ip", name)

    def find_subscription(self, name=None):
        return self._find_resource(self.list_subscriptions, "subscription", name)

    def open_tty(self, name):
        return self.server.open_console(self.find_server(name)["uuid"])

    def close_tty(self, name):
        return self.server.close_console(self.find_server(name)["uuid"])

    def open_vnc(self, name):
        return self.server.open_vnc(self.find_server(name)["uuid"])

    def close_vnc(self, name):
        return self.server.close_vnc(self.find_server(name)["uuid"])

    def convert_memory_value(self, value):
        if value[-1] in ("t", "T"):
            value = float(value[:-1]) * 1024 ** 4
        if value[-1] in ("g", "G"):
            value = float(value[:-1]) * 1024 ** 3
        elif value[-1] in ("m", "M"):
            value = float(value[:-1]) * 1024 ** 2
        elif value[-1] in ("k", "K"):
            value = float(value[:-1]) * 1024
        else:
            value = int(value)
        return int(value)

    def format_memory_value(self, value):
        value = float(value)
        if value >= 1024 ** 4:
            value /= 1024 ** 4
            suffix = "T"
        elif value >= 1024 ** 3:
            value /= 1024 ** 3
            suffix = "G"
        elif value >= 1024 ** 2:
            value /= 1024 ** 2
            suffix = "M"
        elif value >= 1024:
            value /= 1024
            suffix = "K"
        else:
            suffix = ""
        number = "%.1f" % (value)
        if number[-2:] == ".0":
            number = number[:-2]
        return number + suffix

    def map_storage_type(self, storage_type):
        if storage_type == "ssd":
            return "dssd"
        elif storage_type == "magnetic":
            return "zadara"

        raise ParameterError(f"unknown storage_type {storage_type}")

    def create_drive(self, name, size, media, multimount, storage_type):
        return self.drive.create(
            dict(
                name=name,
                size=self.convert_memory_value(size),
                media=media,
                storage_type=self.map_storage_type(storage_type),
                allow_multimount=multimount,
            )
        )

    def modify_drive(
        self, name, rename=None, media=None, multimount=None, storage_type=None
    ):
        drive = self.find_drive(name)
        if rename:
            drive["name"] = rename
        if media:
            drive["media"] = media
        if multimount:
            drive["allow_multimount"] = multimount == "enable"
        if storage_type:
            drive["storage_type"] = storage_type
        return self.drive.update(drive["uuid"], drive)

    def resize_drive(self, drive, size):
        drive["size"] = self.convert_memory_value(size)
        return self.drive.resize(drive["uuid"], drive)

    def create_server(
        self,
        name,
        cpu_count,
        cpu_speed,
        memory,
        password,
        attach_drive,
        create_drive,
        boot_cdrom,
        smp,
    ):
        """create a server, attaching or creating a drive, attaching a boot iso"""

        parameters = dict(
            name=name,
            cpu=cpu_count * cpu_speed,
            smp=cpu_count,
            mem=self.convert_memory_value(memory),
            vnc_password=password,
            cpus_instead_of_cores=bool(smp == "cpu"),
        )
        server = self.server.create(parameters)

        server.setdefault("drives", [])
        if boot_cdrom:
            cdrom = self.find_drive(boot_cdrom)
            if cdrom:
                if cdrom["media"] == "cdrom":
                    server["drives"].append(
                        dict(
                            boot_order=2,
                            dev_channel="0:0",
                            device="ide",
                            drive=cdrom["uuid"],
                        )
                    )
                else:
                    raise ParameterError(
                        f"failed boot cdrom attach; {boot_cdrom} media must be cdrom"
                    )
            else:
                raise ResourceNotFound(
                    f"failed boot cdrom attach; {boot_cdrom} not found"
                )

        if attach_drive:
            drive = self.find_drive(attach_drive)
            if drive["media"] != "disk":
                raise ParameterError(
                    f"failed drive attach; {attach_drive} must be a disk drive"
                )
            elif drive["status"] != "unmounted":
                raise ParameterError(
                    f"failed drive attach; {attach_drive} must be unmounted"
                )
        elif create_drive:
            drive = self.create_drive(
                f"{name}-system", create_drive, "disk", False, "ssd"
            )
        else:
            drive = None

        if drive:
            server["drives"].append(
                dict(
                    boot_order=1,
                    dev_channel="0:0",
                    device="virtio",
                    drive=drive["uuid"],
                )
            )

        # default nic creation is a single public DHCP
        server["nics"] = [
            {
                "ip_v4_conf": {"conf": "dhcp", "ip": None},
                "model": "virtio",
                "vlan": None,
            }
        ]

        return self.server.update(server["uuid"], server)

    def upload_drive_image(self, input_file):
        """upload an image, creating a new drive, and return UUID"""
        s = requests.Session()
        s.auth = (self.config.get("username"), self.config.get("password"))
        s.headers.update({"Content-Type": "application/octet-stream"})
        r = s.post(self.upload_endpoint, data=input_file)
        return r.text.strip()



#!/usr/bin/env python

"""conftest.py"""

import os
from pprint import pprint

import pytest
from click.testing import CliRunner

from cscli import cli

# tests use pytest-vcr to record/replay API interaction
# to perform live tests against the API, valid credentials are required
# use --vcr-record=all to create new recordings
# use --vcr-record=none to use stored recordings
# when using recordings, these environment vars are set
DUMMY_API_AUTH = {
    "CLOUDSIGMA_USERNAME": "user@example.org",
    "CLOUDSIGMA_PASSWORD": "$uper$ecret",
    "CLOUDSIGMA_REGION": "sjc",
}


@pytest.fixture(scope="session", autouse=True)
def api_config_vars():
    for var, value in DUMMY_API_AUTH.items():
        if var not in os.environ:
            os.environ[var] = value


# don't save password in pytest-vcr recordings
@pytest.fixture(scope="module")
def vcr_config():
    return {
        "filter_headers": [("Authorization", "DUMMY")],
    }


class FixtureRunner:
    def __init__(self):
        self.runner = None

    def invoke(self, *args, **kwargs):
        expected_exit_code = kwargs.get("expected_exit_code", 0)
        expected_exception = kwargs.get("raises", None)

        assert isinstance(args, (list, tuple))
        print(f"invoking: {repr(args)}")
        args = list(args)
        args.insert(0, "--debug")
        ret = self.runner.invoke(cli.cli, args)
        pprint(ret)
        if expected_exception:
            assert ret.exception
            assert isinstance(ret.exception, expected_exception)
        else:
            assert ret.exit_code == expected_exit_code
        return ret

    def __enter__(self):
        self.runner = CliRunner()
        print(f"\nstartup {self.runner}")
        return self.invoke

    def __exit__(self, *args, **kwargs):
        print(f"\nteardown {self.runner}")
        self.runner = None
        return False


@pytest.fixture
def runner():
    with FixtureRunner() as _runner:
        yield _runner



#!/usr/bin/env python3

import click

from cscli.cli import pass_environment
from cscli.error import ParameterError


@click.command("list", short_help="list resources by type")
@click.argument(
    "resource",
    type=click.Choice(
        ["servers", "drives", "vlans", "ips", "subscriptions", "capabilities"]
    ),
    metavar="RESOURCE_TYPE",
    required=False,
    default=None,
)
@click.option(
    "fmt",
    "--detail",
    flag_value="detail",
    help="detailed JSON output (default)",
)
@click.option(
    "fmt",
    "--json",
    flag_value="detail",
    hidden=True,
    help="detailed JSON output (default)",
)
@click.option("fmt", "--uuid", flag_value="uuid", help="output resource UUIDs")
@click.option(
    "fmt",
    "--brief",
    flag_value="brief",
    help="abbreviated JSON output",
)
@click.option(
    "fmt",
    "--text",
    flag_value="text",
    help="text output",
)
@pass_environment
def cli(ctx, resource, fmt):
    """list resources: servers drives ips venvs capabilities subscriptions"""
    list_map = {
        "servers": ctx.api.list_servers,
        "drives": ctx.api.list_drives,
        "vlans": ctx.api.list_vlans,
        "ips": ctx.api.list_ips,
        "subscriptions": ctx.api.list_subscriptions,
        "capabilities": ctx.api.list_capabilities,
        None: ctx.api.list_all,
    }

    if resource == "capabilities" and fmt not in [None, "detail"]:
        raise ParameterError(f"{resource} cannot be formatted as {fmt}")

    if resource == "subscriptions" and fmt not in [None, "uuid", "detail"]:
        raise ParameterError(f"{resource} cannot be formatted as {fmt}")

    ret = list_map[resource](fmt)
    if fmt == "text":
        click.echo("=" * 79)
        for name, data in ret.items():
            first = True
            for result in data:
                for uuid, lines in result.items():
                    if first:
                        first = False
                    else:
                        click.echo("-" * 79)
                    click.echo(f"{name[:-1]} {uuid}")
                    for line in lines:
                        click.echo("   " + line)
            click.echo("=" * 79)
    elif fmt == "uuid":
        uuids = {}
        for name, data in ret.items():
            uuids[name] = [u for u in [list(d.keys())[0] for d in data] if u]
        ctx.output(uuids)
    else:
        ctx.output(ret)



import json
import os
import sys

import click
import yaml

from cscli import CloudSigmaClient, __description__, __version__

CONTEXT_SETTINGS = dict(auto_envvar_prefix="CSCLI")


class Environment:
    def __init__(self):
        self.verbose = False
        self.compact = False
        self.fmt = "json"
        self.api = None

    def log(self, msg, *args):
        """Logs a message to stderr."""
        if self.verbose:
            if args:
                msg %= args
            click.echo(msg, file=sys.stderr)

    def output(self, item, status=True):
        if self.compact:
            json_indent = None
            json_separators = [",", ":"]
        else:
            json_indent = 2
            json_separators = [", ", ": "]
        if self.fmt == "yaml":
            out = yaml.dump(item)
        else:
            out = json.dumps(
                dict(status=status, result=item),
                indent=json_indent,
                separators=json_separators,
            )
        click.echo(out)

    def error(self, message):
        self.output(message, False)
        sys.exit()

    def confirm(self, resource, action, label, force):
        if not force:
            if not click.confirm(
                f"Confirm {action} of {label} {resource['name']} {resource['uuid']}"
            ):
                self.output(f"{action} averted")
                sys.exit()


pass_environment = click.make_pass_decorator(Environment, ensure=True)
cmd_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "commands"))


class ComplexCLI(click.MultiCommand):
    def list_commands(self, ctx):
        rv = []
        for filename in os.listdir(cmd_folder):
            if filename.endswith(".py") and filename.startswith("cmd_"):
                print(f"filename={filename}")
                rv.append(filename[4:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        try:
            mod = __import__(f"cscli.commands.cmd_{name}", None, None, ["cli"])
        except ImportError:
            return
        return mod.cli


@click.command("cscli", cls=ComplexCLI, context_settings=CONTEXT_SETTINGS)
@click.version_option(message=f"{__package__} v{__version__} -- {__description__}")
@click.option(
    "-u", "--username", type=str, help="override env var CLOUDSIGMA_USERNAME]"
)
@click.option(
    "-p", "--password", type=str, help="override env var CLOUDSIGMA_PASSWORD]"
)
@click.option("-r", "--region", type=str, help="override env var CLOUDSIGMA_REGION]")
@click.option(
    "-d", "--debug", is_flag=True, help="output full stacktrace on runtime error"
)
@click.option("-v", "--verbose", is_flag=True, help="Enables verbose mode.")
@click.option("-c", "--compact", is_flag=True, help="Output Compact JSON")
@click.option("-y", "--yaml", is_flag=True, help="Output YAML")
@click.option("-j", "--json", is_flag=True, help="Output JSON")
@pass_environment
def cli(ctx, region, username, password, debug, verbose, compact, yaml, json):
    """CLI for the CloudSigma API

    create, modify, operate, and destroy resources on cloudsigma
    """

    def exception_handler(
        exception_type, exception, traceback, debug_hook=sys.excepthook
    ):
        if debug:
            debug_hook(exception_type, exception, traceback)
        else:
            ctx.error(f"{exception_type.__name__}: {exception}")

    sys.excepthook = exception_handler

    ctx.verbose = verbose
    ctx.compact = compact
    if yaml:
        ctx.fmt = "yaml"
    if json:
        ctx.fmt = "json"

    ctx.api = CloudSigmaClient(region, username, password)



import secrets
import string
import subprocess
import time

import click

from cscli import MIN_CPU, MIN_DISK, MIN_MHZ, MIN_RAM, PASSWORD_LEN
from cscli.cli import pass_environment


def mkpasswd(length):
    codex = (
        string.ascii_lowercase
        + string.ascii_uppercase
        + string.digits
        + "!#%()*+,-./:;<=>@^_"
    )
    return "".join([secrets.choice(codex) for _ in range(length)])


@click.group(name="server")
@click.argument("name", type=str, metavar="SERVER_NAME_OR_UUID")
@pass_environment
def cli(ctx, name):
    """server actions: create list show destroy attach detach start stop ttyopen, ttyclose, shutdown"""
    ctx.server_name = name


@cli.command()
@click.option(
    "-c", "--cpu", metavar="CPU_COUNT", type=int, default=MIN_CPU, help="SMP CPU count"
)
@click.option(
    "-s", "--speed", metavar="CPU_MHZ", type=int, default=MIN_MHZ, help="Mhz per cpu"
)
@click.option(
    "-m",
    "--memory",
    metavar="SIZE",
    type=str,
    default=MIN_RAM,
    help="memory (supports M or G suffix)",
)
@click.option("-p", "--password", type=str, default=mkpasswd(PASSWORD_LEN))
@click.option(
    "-a",
    "--attach-drive",
    metavar="NAME",
    type=str,
    default=None,
    help="drive name or uuid to attach",
)
@click.option(
    "-d",
    "--create-drive",
    metavar="SIZE",
    type=str,
    default=MIN_DISK,
    help="drive size to create",
)
@click.option(
    "-b", "--boot-cdrom", metavar="NAME", type=str, help="boot cdrom drive name or uuid"
)
@click.option(
    "-S",
    "--smp",
    type=click.Choice(["cpu", "core"]),
    help="smp as multi-core or multi-cpu",
)
@pass_environment
def create(
    ctx, cpu, speed, memory, password, attach_drive, create_drive, boot_cdrom, smp
):
    """create a server resource, optionally create or attach a drive and/or boot cdrom"""
    ctx.output(
        ctx.api.create_server(
            ctx.server_name,
            cpu,
            speed,
            memory,
            password,
            attach_drive,
            create_drive,
            boot_cdrom,
            smp,
        )
    )


@cli.command()
@click.option("-k", "--keep-drives", is_flag=True)
@click.option("-f", "--force", is_flag=True, help="suppress confirmation prompt")
@pass_environment
def destroy(ctx, keep_drives, force):
    """delete server, optionally preserve drives, always preserve cdroms"""
    server = ctx.api.find_server(ctx.server_name)
    uuid = server["uuid"]
    ctx.confirm(server, "destruction", "server", force)
    if keep_drives:
        drives = None
    else:
        drives = "disks"
    ctx.output(
        ctx.api.server.delete(server["uuid"], drives)
        or f"server {uuid} '{ctx.server_name}' destroyed"
    )


@cli.command()
@pass_environment
def start(ctx):
    """power on"""
    ctx.output(ctx.api.server.start(ctx.api.find_server(ctx.server_name)["uuid"]))


@cli.command()
@pass_environment
def stop(ctx):
    """power off"""
    ctx.output(ctx.api.server.stop(ctx.api.find_server(ctx.server_name)["uuid"]))


@cli.command()
@pass_environment
def shutdown(ctx):
    """ACPI shutdown (soft power switch)"""
    ctx.output(ctx.api.server.shutdown(ctx.api.find_server(ctx.server_name)["uuid"]))


@cli.command()
@pass_environment
def show(ctx):
    """display server attributes"""
    ctx.output(ctx.api.find_server(ctx.server_name))


@cli.command()
@pass_environment
def runtime(ctx):
    """display server runtime data"""
    ctx.output(ctx.api.server.runtime(ctx.api.find_server(ctx.server_name)["uuid"]))


@cli.command()
@click.option("-c", "--close", is_flag=True, help="close the TTY port")
@click.option(
    "-e",
    "--exec",
    "exec_command",
    type=str,
    metavar="COMMAND",
    help="execute COMMAND ADDRESS PORT",
)
@click.option(
    "-p",
    "--password",
    is_flag=True,
    help="pipe vnc_password to stdin of --exec command",
)
@pass_environment
def tty(ctx, close, exec_command, password):
    """open or close a console TTY port"""
    if close:
        ctx.output(ctx.api.close_tty(ctx.server_name))
    else:
        ret = ctx.api.open_tty(ctx.server_name)
        if ret:
            if exec_command:
                if password:
                    server = ctx.api.find_server(ctx.server_name)
                    pipe = f"echo '{server['vnc_password']}'|"
                else:
                    pipe = ""
                url = ret["console_url"]
                addr, port = (url.split("/")[2]).split(":")
                cmd = f"{pipe}{exec_command} {addr} {port}"
                exit_code = subprocess.call(cmd, shell=True)
                ret = f"subprocess returned {exit_code}"
        ctx.output(ret)


@cli.command()
@click.option("-c", "--close", is_flag=True)
@pass_environment
def vnc(ctx, close):
    """open or close a VNC session"""
    if close:
        ctx.output(ctx.api.close_vnc(ctx.server_name))
    else:
        ctx.output(ctx.api.open_vnc(ctx.server_name))


@cli.command()
@click.option(
    "-s", "--status", type=str, default="started", help="desired server status"
)
@click.option(
    "-t", "--timeout", type=int, default=15, help="timeout in seconds, 0=infinite"
)
@pass_environment
def wait(ctx, status, timeout):
    """wait for server status"""
    server = ctx.api.find_server(ctx.server_name)
    if timeout:
        timeout = time.time() + timeout
    while server["status"] != status:
        server = ctx.api.find_server(ctx.server_name)
        if timeout and time.time() > timeout:
            ctx.error(f"Timeout waiting for server {ctx.server_name} status {status}")
    ctx.output({"uuid": server["uuid"], "status": server["status"]})


@cli.command()
@click.argument("drive", type=str)
@click.option("-c", "--dev-channel", type=str, default="0:0")
@click.option("-d", "--device", type=click.Choice(["virtio", "ide"]), default="virtio")
@pass_environment
def attach(ctx, drive, dev_channel, device):
    """attach drive to server"""
    server = ctx.api.find_server(ctx.server_name)
    attachment = dict(
        drive=ctx.api.find_drive(drive),
        dev_channel=dev_channel,
        device=device,
        boot_order=len(server["drives"]) + 1,
    )
    server.setdefault("drives", []).append(attachment)
    ctx.output(ctx.api.server.update(server["uuid"], server))


@cli.command()
@click.argument("drive", metavar="DRIVE", type=str)
@pass_environment
def detach(ctx, drive):
    """detach drive from server"""
    server = ctx.api.find_server(ctx.server_name)
    drive_uuid = ctx.api.find_drive(drive)["uuid"]
    found = False
    for index, attached_drive in enumerate(server["drives"]):
        if drive_uuid == attached_drive["drive"]["uuid"]:
            del server["drives"][index]
            found = True
            break
    if found:
        ctx.output(ctx.api.server.update(server["uuid"], server))
    else:
        ctx.error(f"server {ctx.server_name} has no attached drive {drive}")


@cli.command()
@click.argument("action", type=click.Choice(["add", "delete", "modify"]), default="add")
@click.option(
    "-c",
    "--config",
    type=click.Choice(["dhcp", "static", "manual", "vlan"]),
    default="dhcp",
)
@click.option("-m", "--model", type=str)
@click.option("-i", "--ip", type=str)
@click.option("-m", "--mac", type=str)
@click.option("--vlan", type=str)
@pass_environment
def nic(ctx, action, config, model, ip, mac, vlan):
    """add, delete, or modify network interfaces"""
    server = ctx.api.find_server(ctx.server_name)
    server.setdefault("nics", [])

    new_nic = {}

    if model:
        new_nic["model"] = model

    if config == "dhcp":
        new_nic["ip_v4_conf"] = dict(conf="dhcp")
    elif config == "static":
        if not ip:
            ctx.error("static config requires --ip option (subscribed IP)")
        # TODO: verify the ip is subscribed and not already assigned
        new_nic["ip_v4_conf"] = dict(conf="static", ip=ip)
    elif config == "manual":
        new_nic["ip_v4_conf"] = dict(conf="manual")
    elif config == "vlan":
        if not vlan:
            ctx.error("vlan config requires --vlan option (subscribed VLAN)")
        # TODO: verify the VLAN is subscribed
        new_nic["vlan"] = vlan
    else:
        ctx.error(f"unknown config type {config}")

    if action == "add":
        server["nics"].append(new_nic)
    else:
        if not mac:
            ctx.error(f"{action} action requires --mac option")

        # find index and nic of server.nics with matching mac address
        matches = [(i, n) for (i, n) in enumerate(server["nics"]) if n["mac"] == mac]
        if len(matches) != 1:
            ctx.error(f"{action} mac addr {mac} not found on server {ctx.server_name}")
        nic_index = matches[0][0]
        if action == "delete":
            del server["nics"][nic_index]
        elif action == "modify":
            new_nic["mac"] = mac
            server["nics"][nic_index] = new_nic
        else:
            ctx.error(f"unknown action {action}")
    ctx.output(ctx.api.server.update(server["uuid"], server))


@cli.command()
@click.option("-r", "--rename", type=str)
@click.option(
    "-c", "--cpu", metavar="CPU_COUNT", type=int, default=MIN_CPU, help="SMP CPU count"
)
@click.option(
    "-s", "--speed", metavar="CPU_MHZ", type=int, default=MIN_MHZ, help="Mhz per cpu"
)
@click.option("-m", "--memory", type=str)
@click.option("-p", "--password", type=str)
@click.option(
    "-S",
    "--smp",
    type=click.Choice(["cpu", "core"]),
    help="smp as multi-core or multi-cpu",
)
@pass_environment
def modify(ctx, rename, cpu, speed, memory, password, smp):
    """modify server attributes"""

    server = ctx.api.find_server(ctx.server_name)
    if rename:
        server["name"] = rename
    if cpu or speed:
        server["cpu"] = cpu * speed
        server["smp"] = cpu
    if memory:
        server["mem"] = ctx.api.convert_memory_value(memory)
    if password:
        server["vnc_password"] = password
    if smp:
        server["cpus_instead_of_cores"] = bool(smp == "cpu")
    ctx.output(ctx.api.server.update(server["uuid"], server))



"""Top-level package for cscli."""

__author__ = """Matt Krueger"""
__email__ = "mkrueger@rstms.net"
__version__ = "0.3.2"
__description__ = "CloudSigma API command line interface"
__timestamp__ = "2021-09-16T02:07:07-05:00"

from .api_client import CloudSigmaClient

MIN_CPU = 1
MIN_MHZ = 1000
MIN_RAM = "256M"
MIN_DISK = "512M"

PASSWORD_LEN = 24

__all__ = [CloudSigmaClient]



#!/usr/bin/env python

import pytest

"""Tests for `cscli` package."""


@pytest.mark.vcr()
def test_cli_init(runner):
    """Test the CLI."""
    result = runner(expected_exit_code=2)
    assert "Error: Missing command." in result.output


@pytest.mark.vcr()
def test_cli_help(runner):
    """Test top-level --help"""
    help_result = runner("--help")
    assert "Usage: cscli [OPTIONS] COMMAND [ARGS]..." in help_result.output



from pprint import pprint

import pytest

from cscli.api_client import CloudSigmaClient


@pytest.fixture()
def api():
    return CloudSigmaClient()


def test_client_init(api):
    assert api


all_keys = ["servers", "drives", "ips", "vlans", "capabilities", "subscriptions"]


def _verify(ret, keys):
    pprint(ret)
    assert ret
    assert isinstance(ret, dict)
    for key, value in ret.items():
        assert isinstance(key, str)
        assert isinstance(value, list)
        assert key in keys
        for resource in value:
            assert isinstance(resource, dict)


@pytest.mark.vcr()
def test_client_list_servers_none(api):
    _verify(api.list_servers(None), ["servers"])


@pytest.mark.vcr()
def test_client_list_drives_none(api):
    _verify(api.list_drives(None), ["drives"])


@pytest.mark.vcr()
def test_client_list_ips_none(api):
    _verify(api.list_ips(None), ["ips"])


@pytest.mark.vcr()
def test_client_list_vlans_none(api):
    _verify(api.list_vlans(None), ["vlans"])


@pytest.mark.vcr()
def test_client_list_capabilities_none(api):
    _verify(api.list_capabilities(None), ["capabilities"])


@pytest.mark.vcr()
def test_client_list_subscriptions_none(api):
    _verify(api.list_subscriptions(None), ["subscriptions"])


@pytest.mark.vcr()
def test_client_list_all_none(api):
    _verify(api.list_all(None), all_keys)


@pytest.mark.vcr()
def test_client_list_servers_detail(api):
    _verify(api.list_servers("detail"), ["servers"])


@pytest.mark.vcr()
def test_client_list_drives_detail(api):
    _verify(api.list_drives("detail"), ["drives"])


@pytest.mark.vcr()
def test_client_list_ips_detail(api):
    _verify(api.list_ips("detail"), ["ips"])


@pytest.mark.vcr()
def test_client_list_vlans_detail(api):
    _verify(api.list_vlans("detail"), ["vlans"])


@pytest.mark.vcr()
def test_client_list_capabilities_detail(api):
    _verify(api.list_capabilities("detail"), ["capabilities"])


@pytest.mark.vcr()
def test_client_list_subscriptions_detail(api):
    _verify(api.list_subscriptions("detail"), ["subscriptions"])


@pytest.mark.vcr()
def test_client_list_all_detail(api):
    _verify(api.list_all("detail"), all_keys)


@pytest.mark.vcr()
def test_client_list_servers_brief(api):
    _verify(api.list_servers("brief"), ["servers"])


@pytest.mark.vcr()
def test_client_list_drives_brief(api):
    _verify(api.list_drives("brief"), ["drives"])


@pytest.mark.vcr()
def test_client_list_ips_brief(api):
    _verify(api.list_ips("brief"), ["ips"])


@pytest.mark.vcr()
def test_client_list_vlans_brief(api):
    _verify(api.list_vlans("brief"), ["vlans"])


@pytest.mark.vcr()
def test_client_list_capabilities_brief(api):
    _verify(api.list_capabilities("brief"), ["capabilities"])


@pytest.mark.vcr()
def test_client_list_subscriptions_brief(api):
    _verify(api.list_subscriptions("brief"), ["subscriptions"])


@pytest.mark.vcr()
def test_client_list_all_brief(api):
    _verify(api.list_all("brief"), all_keys)


@pytest.mark.vcr()
def test_client_list_servers_uuid(api):
    _verify(api.list_servers("uuid"), ["servers"])


@pytest.mark.vcr()
def test_client_list_drives_uuid(api):
    _verify(api.list_drives("uuid"), ["drives"])


@pytest.mark.vcr()
def test_client_list_ips_uuid(api):
    _verify(api.list_ips("uuid"), ["ips"])


@pytest.mark.vcr()
def test_client_list_vlans_uuid(api):
    _verify(api.list_vlans("uuid"), ["vlans"])


@pytest.mark.vcr()
def test_client_list_capabilities_uuid(api):
    _verify(api.list_capabilities("uuid"), ["capabilities"])


@pytest.mark.vcr()
def test_client_list_subscriptions_uuid(api):
    _verify(api.list_subscriptions("uuid"), ["subscriptions"])


@pytest.mark.vcr()
def test_client_list_all_uuid(api):
    _verify(api.list_all("uuid"), all_keys)


@pytest.mark.vcr()
def test_client_list_servers_text(api):
    _verify(api.list_servers("text"), ["servers"])


@pytest.mark.vcr()
def test_client_list_drives_text(api):
    _verify(api.list_drives("text"), ["drives"])


@pytest.mark.vcr()
def test_client_list_ips_text(api):
    _verify(api.list_ips("text"), ["ips"])


@pytest.mark.vcr()
def test_client_list_vlans_text(api):
    _verify(api.list_vlans("text"), ["vlans"])


@pytest.mark.vcr()
def test_client_list_capabilities_text(api):
    _verify(api.list_capabilities("text"), ["capabilities"])


@pytest.mark.vcr()
def test_client_list_subscriptions_text(api):
    _verify(api.list_subscriptions("text"), ["subscriptions"])


@pytest.mark.vcr()
def test_client_list_all_text(api):
    _verify(api.list_all("text"), all_keys)



#!/usr/bin/env python

"""Tests for command/cmd_list.py module"""
import pytest

from cscli.error import ParameterError


@pytest.mark.vcr()
def test_list_help(runner):
    """Test cscli list --help"""
    result = runner("--debug", "list", "--help")
    assert "Usage: cscli list" in result.output


@pytest.mark.vcr()
def test_list_all_default(runner):
    runner("list")


@pytest.mark.vcr()
def test_list_all_detail(runner):
    runner("list", "--detail")


@pytest.mark.vcr()
def test_list_all_uuid(runner):
    runner("list", "--uuid")


@pytest.mark.vcr()
def test_list_all_brief(runner):
    runner("list", "--brief")


@pytest.mark.vcr()
def test_list_all_text(runner):
    runner("list", "--text")


@pytest.mark.vcr()
def test_list_servers_default(runner):
    runner("list", "servers")


@pytest.mark.vcr()
def test_list_servers_detail(runner):
    runner("list", "--detail", "servers")


@pytest.mark.vcr()
def test_list_servers_uuid(runner):
    runner("list", "--uuid", "servers")


@pytest.mark.vcr()
def test_list_servers_brief(runner):
    runner("list", "--brief", "servers")


@pytest.mark.vcr()
def test_list_servers_text(runner):
    runner("list", "--text", "servers")


@pytest.mark.vcr()
def test_list_drives_default(runner):
    runner("list", "drives")


@pytest.mark.vcr()
def test_list_drives_detail(runner):
    runner("list", "--detail", "drives")


@pytest.mark.vcr()
def test_list_drives_uuid(runner):
    runner("list", "--uuid", "drives")


@pytest.mark.vcr()
def test_list_drives_brief(runner):
    runner("list", "--brief", "drives")


@pytest.mark.vcr()
def test_list_drives_text(runner):
    runner("list", "--text", "drives")


@pytest.mark.vcr()
def test_list_ips_default(runner):
    runner("list", "ips")


@pytest.mark.vcr()
def test_list_ips_detail(runner):
    runner("list", "--detail", "ips")


@pytest.mark.vcr()
def test_list_ips_uuid(runner):
    runner("list", "--uuid", "ips")


@pytest.mark.vcr()
def test_list_ips_brief(runner):
    runner("list", "--brief", "ips")


@pytest.mark.vcr()
def test_list_ips_text(runner):
    runner("list", "--text", "ips")


@pytest.mark.vcr()
def test_list_vlans_default(runner):
    runner("list", "vlans")


@pytest.mark.vcr()
def test_list_vlans_detail(runner):
    runner("list", "--detail", "vlans")


@pytest.mark.vcr()
def test_list_vlans_uuid(runner):
    runner("list", "--uuid", "vlans")


@pytest.mark.vcr()
def test_list_vlans_brief(runner):
    runner("list", "--brief", "vlans")


@pytest.mark.vcr()
def test_list_vlans_text(runner):
    runner("list", "--text", "vlans")


@pytest.mark.vcr()
def test_list_subscriptions_default(runner):
    runner("list", "subscriptions")


@pytest.mark.vcr()
def test_list_subscriptions_detail(runner):
    runner("list", "--detail", "subscriptions")


@pytest.mark.vcr()
def test_list_subscriptions_uuid(runner):
    runner("list", "--uuid", "subscriptions")


@pytest.mark.vcr()
def test_list_subscriptions_brief(runner):
    runner("list", "--brief", "subscriptions", raises=ParameterError)


@pytest.mark.vcr()
def test_list_subscriptions_text(runner):
    runner("list", "--text", "subscriptions", raises=ParameterError)


@pytest.mark.vcr()
def test_list_capabilities_default(runner):
    runner("list", "capabilities")


@pytest.mark.vcr()
def test_list_capabilities_detail(runner):
    runner("list", "--detail", "capabilities")


@pytest.mark.vcr()
def test_list_capabilities_uuid(runner):
    runner("list", "--uuid", "capabilities", raises=ParameterError)


@pytest.mark.vcr()
def test_list_capabilities_brief(runner):
    runner("list", "--brief", "capabilities", raises=ParameterError)


@pytest.mark.vcr()
def test_list_capabilities_text(runner):
    runner("list", "--text", "capabilities", raises=ParameterError)



#!/usr/bin/env python3

import click

from cscli import MIN_DISK
from cscli.cli import pass_environment


@click.group("drive", short_help="manage drives")
@click.argument("drive", metavar="NAME_OR_UUID", type=str)
@pass_environment
def cli(ctx, drive):
    """actions: create destroy list show modify snapshot upload download"""
    ctx.drive_name = drive


@cli.command()
@click.option("-s", "--size", type=str, default=MIN_DISK)
@click.option("-m", "--media", type=click.Choice(["disk", "cdrom"]), default="disk")
@click.option(
    "-M", "--multimount", type=click.Choice(["enable", "disable"]), default="disable"
)
@click.option(
    "-t", "--storage-type", type=click.Choice(["ssd", "magnetic"]), default="ssd"
)
@pass_environment
def create(ctx, size, media, multimount, storage_type):
    """create a drive resource"""
    ctx.output(
        ctx.api.create_drive(
            ctx.drive_name, size, media, multimount == "enable", storage_type
        )
    )


@cli.command()
@pass_environment
def show(ctx):
    """display drive by name or uuid"""
    ctx.output(ctx.api.find_drive(ctx.drive_name))


@cli.command()
@click.option("-f", "--force", is_flag=True, help="suppress confirmation prompt")
@pass_environment
def destroy(ctx, force):
    """delete drive by name or uuid"""
    drive = ctx.api.find_drive(ctx.drive_name)
    uuid = drive["uuid"]
    ctx.confirm(drive, "destruction", "drive", force)
    ctx.output(
        ctx.api.drive.delete(drive["uuid"])
        or f"drive {uuid} '{ctx.drive_name}' destroyed"
    )


@cli.command()
@pass_environment
def snapshot(ctx):
    """create drive snapshot"""
    ctx.output(ctx.api.server.snapshot(ctx.api.find_drive(ctx.drive_name)))


@cli.command()
@click.option("-r", "--rename", type=str)
@click.option("-m", "--media", type=click.Choice(["disk", "cdrom"]))
@click.option("-M", "--multimount", type=click.Choice(["enable", "disable"]))
@click.option("-s", "--storage-type", type=click.Choice(["ssd", "magnetic"]))
@pass_environment
def modify(ctx, rename, media, multimount, storage_type):
    """change drive characteristics"""
    ctx.output(
        ctx.api.modify_drive(ctx.drive_name, rename, media, multimount, storage_type)
    )


@cli.command()
@click.option("-s", "--size", type=str, required=True)
@click.option("-f", "--force", is_flag=True, help="suppress confirmation prompt")
@pass_environment
def resize(ctx, size, force):
    """resize an unmounted drive"""
    drive = ctx.api.find_drive(ctx.drive_name)
    ctx.confirm(drive, "resize", "drive", force)
    ctx.output(ctx.api.resize_drive(drive, size))


@cli.command()
@click.argument("input", metavar="DRIVE_IMAGE", type=click.File("rb"))
@click.option("-m", "--media", type=click.Choice(["disk", "cdrom"]), default="disk")
@click.option(
    "-M", "--multimount", type=click.Choice(["enable", "disable"]), default="disable"
)
@pass_environment
def upload(ctx, input, media, multimount):
    """upload a binary disk image or ISO file"""
    uuid = ctx.api.upload_drive_image(input)
    drive = ctx.api.find_drive(uuid)
    ctx.output(
        ctx.api.modify_drive(drive["uuid"], ctx.drive_name, media, multimount, None)
    )


@cli.command()
@click.argument("image-file", type=click.File("wb"))
@pass_environment
def download(ctx, image_file):
    """download drive image to local file"""
    ctx.output(ctx.api.download(ctx.drive_name, image_file))



#!/usr/bin/env python3


class CloudSigmaClientError(Exception):
    def __init__(self, message):
        self.message = message


class ParameterError(CloudSigmaClientError):
    pass


class ResourceNotFound(CloudSigmaClientError):
    pass



#!/usr/bin/env python3

import click

from cscli.cli import pass_environment


@click.group(name="ip", short_help="manage ip addresses")
@click.argument("name", metavar="IP", type=str)
@pass_environment
def cli(ctx, name):
    """IP commands: create list show modify destroy"""
    ctx.ip_name = name


@cli.command()
@click.option("-d", "--duration", type=int, default=1)
@click.option(
    "-u", "--units", type=click.Choice(["hour", "day", "month", "year"]), default="day"
)
@click.option("-a", "--auto-renew", is_flag=True, envvar="CCS_AUTORENEW")
@pass_environment
def create(ctx, duration, units, auto_renew):
    """subscribe to a public IP"""
    if duration > 1:
        units += "s"
    ctx.output(ctx.api.create_ip(ctx.ip_name, f"{duration} {units}", auto_renew))


@cli.command()
@pass_environment
def show(ctx):
    """display IP attributes"""
    ctx.output(ctx.api.find_ip(ctx.ip_name))


@cli.command()
@click.option("-r", "--rename", type=str)
@click.option("-D", "--description", type=str)
# @click.option('-d', '--duration', type=int, default=1)
# @click.option('-u', '--units', type=click.Choice(['hour', 'day', 'month', 'year']), default='day')
# @click.option('-a', '--auto-renew', is_flag=True)
@pass_environment
def modify(ctx, rename, description):
    """modify IP attributes"""
    ip = ctx.api.find_ip(ctx.ip_name)
    if rename:
        ip["meta"]["name"] = rename
    if description:
        ip["meta"]["description"] = description
    # if auto_renew:
    #    ctx.error('unimplemented')
    # if duration:
    #    ctx.error('unimplemented')
    # if units:
    #    ctx.error('unimplemented')
    ctx.output(ctx.api.ip.update(ip["uuid"], ip))


@cli.command()
@click.option("-f", "--force", is_flag=True, help="suppress confirmation prompt")
@pass_environment
def destroy(ctx, force):
    """delete IP"""
    name = ctx.ip_name
    ip = ctx.api.find_ip(name)
    ctx.confirm(ip, "destruction", "IP", force)
    ctx.error(f"IP {name} cannot be deleted while subscribed (verify autorenew status)")
